/* gnome-url.c
 * Copyright (C) 1998, James Henstridge <james@daa.com.au>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc.,  59 Temple Place - Suite 330, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <string.h>
#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-config.h>
#include <libgnome/gnome-exec.h>
#include <libgnome/gnome-util.h>
#include "gnome-popt.h" /* poptParseArgvString */
#include "gnome-help.h"
#include "gnome-url.h"

static gchar *uri_get_file_uri_from_ghelp_uri (const gchar *path);

#define DEFAULT_HANDLER "xdg-open \"%s\""

static gchar *
gnome_url_default_handler (const char *protocol)
{
	static gchar *default_handler = 0;

	if (!default_handler) {
		gchar *str, *app;
		gboolean def;
		str = gnome_config_get_string_with_default ("/Gnome/URL Handlers/default-show",
							    &def);
                
		if (def) {
			app = gnome_is_program_in_path ("nautilus");
			if (app) {
				g_free (app);
				app = "nautilus \"%s\"";
			} else
				app = "gnome-help-browser \"%s\"";

			/* first time gnome_url_show is run -- set up some useful defaults */
			g_free (gnome_config_get_string_with_default(
				"/Gnome/URL Handlers/info-show", &def));
			if (def)
				gnome_config_set_string ("/Gnome/URL Handlers/info-show", app);
			g_free (gnome_config_get_string_with_default(
				"/Gnome/URL Handlers/man-show", &def));
			if (def)
				gnome_config_set_string ("/Gnome/URL Handlers/man-show", app);
			g_free (gnome_config_get_string_with_default(
				"/Gnome/URL Handlers/ghelp-show", &def));
			if (def) 
				gnome_config_set_string ("/Gnome/URL Handlers/ghelp-show", app);

			gnome_config_sync_file ("/Gnome/");

			if (protocol &&
			    (strcmp (protocol, "ghelp") == 0 ||
			      strcmp (protocol, "info") == 0 ||
			      strcmp (protocol, "man") == 0)) {
				return app;
			} else {
                                /* always force use of xdg-open */
                                default_handler = DEFAULT_HANDLER;
                        }
		} else {
                        if (protocol &&
                            (strcmp (protocol, "ghelp") == 0 ||
                             strcmp (protocol, "info") == 0 ||
                             strcmp (protocol, "man") == 0)) {
                                default_handler = str;
                        } else {
                                /* force xdg-open */
                                default_handler = DEFAULT_HANDLER;
                        }
                }
	}
	return default_handler;
}

/**
 * gnome_url_show
 * @url: URL to show
 *
 * Description:
 * Loads the given URL in an appropriate viewer.  The viewer is deduced from
 * the protocol part of the URL.  That is all that the caller should know
 * about the function.
 *
 * Internally, the handler for a given URL is deduced by looking in the
 * /Gnome/URL Handlers/<protocol>-show key in the user's configuration
 * files.  The key is a string that will be passed to gnome_execute_shell(),
 * after the %s is replaced with with the URL.  If that key can't be found,
 * it falls back to /Gnome/URL Handlers/default-show, and if that isn't
 * found, uses the contents of the DEFAULT_HANDLER macro in this file.
 *
 * If no /Gnome/URL Handlers keys are set, some sensible defaults are added
 * to the user's configuration files.
 *
 * We really need a capplet so a user can configure the behaviour of this
 * function.
 **/
void
gnome_url_show(const gchar *url)
{
	gint i;
	gchar *pos, *template;
	gchar *tmp_url = NULL;
	gboolean free_template = FALSE;
	int argc;
       	char **argv;
	
	g_return_if_fail (url != NULL);
	pos = strchr (url, ':');

	if (pos){
		gchar *protocol, *path;
		gboolean def;
		
		g_return_if_fail (pos >= url);
		protocol = g_new (gchar, pos - url + 1);
		strncpy (protocol, url, pos - url);
		protocol [pos - url] = '\0';
		g_strdown (protocol);

		if (strcmp (protocol, "ghelp") == 0) {
			/* We no longer have any app (other than Galeon)
			 * that can handle both ghelp: URI's and GNOME 1
			 * help files, so do a hardcoded translation
			 * and then run xdg-open.
			 */
			if (strstr (url, "galeon") != NULL) {
				/* But always use Galeon for Galeon */
				template = "galeon \"%s\"";
			} else {
				url = tmp_url = uri_get_file_uri_from_ghelp_uri (pos + 1);
				template = "xdg-open \"%s\"";
			}
		} else {
			path = g_strconcat ("/Gnome/URL Handlers/", protocol, "-show", NULL);
			template = gnome_config_get_string_with_default (path, &def);
			g_free (path);
			if (def)
				template = gnome_url_default_handler (protocol);
			else
				free_template = TRUE;
			g_free (protocol);
		}
	} else /* no : ? -- this shouldn't happen.  Use default handler */
		template = gnome_url_default_handler (NULL);

	/* we use a popt function as it does exactly what we want to do and
	   gnome already uses popt */
	if(poptParseArgvString(template, &argc, &argv) != 0) {
		/* can't parse */
		g_warning("Parse error of '%s'", template);
		return;
	}

	/* we can just replace the entry in the array since the
	 * array is all in one buffer, we won't leak */
	for(i = 0; i < argc; i++) {
		if(strcmp(argv[i], "%s") == 0)
			/* we can safely cast as it will not get freed nor
			 * modified */
			argv[i] = (char *)url;
	}

	/* use execute async, and not the shell, shell is evil and a
	 * security hole */
	gnome_execute_async (NULL, argc, argv);

	if (tmp_url)
		g_free (tmp_url);
	
	if (free_template)
		g_free (template);

	/* the way the poptParseArgvString works is that the entire thing
	 * is allocated as one buffer, so just free will suffice, also
	 * it must be free and not g_free */
	free(argv);
}

/* Next two functions borrowed from yelp, and modified to use
 * gnome_help_file_path().
 *
 * Copyright (C) 2002 Mikael Hallendal <micke@codefactory.se>
 *
 * From: Mikael Hallendal <micke@codefactory.se>
 * To: Owen Taylor <otaylor@redhat.com>
 * Date: 29 Aug 2002 22:29:37 +0200
 * Message-Id: <1030652977.32141.2.camel@fluffy.hallendal.net>
 *
 * Hi!
 * 
 * I give permission to use the code in yelp-uri.c under the LPGL.
 * 
 * Regards,
 *   Mikael Hallendal
 */
static gchar *
uri_get_path_from_relative (const gchar *path)
{
	gchar        *doc_id;
	gchar        *file_name;
	gchar        *ret_val = NULL;
	const gchar  *ch, *hash;
	int len;
	
	if ((ch = strchr (path, '/'))) {
		/* 2:  ghelp:AisleRiot2/Klondike */
		doc_id    = g_strndup (path, ch - path);

		hash = strchr (ch + 1, '#');
		if (!hash)
			hash = ch + 1 + strlen (ch + 1);
		len = hash - (ch + 1);

		if (len >= 5 && strncmp (hash - 5, ".html", 5) == 0)
			file_name = g_strdup (ch + 1);
		else {
			char *tmp = g_strndup (ch + 1, len);
			file_name = g_strconcat (tmp,
						 ".html",
						 *hash ? "#" : NULL,
						 *hash ? hash + 1 : NULL,
						 NULL);
			g_free (tmp);
		}
	} else {
		/* 1:  ghelp:nautilus */
		doc_id    = (gchar *)path;
		file_name = g_strdup ("index.html");
	}

	ret_val = gnome_help_file_path (doc_id, file_name);

	if (doc_id != path)
		g_free (doc_id);
	g_free (file_name);

	return ret_val;
}

static gchar *
uri_get_path_from_ghelp_uri (const gchar *path)
{
	gchar *ret_val = NULL;
	gchar *work_path;
	
	work_path = g_strdup (path);
	
	g_strstrip (work_path);

	if (path[0] == '/') {
		/* Absolute URL */
		gint i = 1;
		gint len = strlen (work_path);
		
		while (i < len && work_path[i] == '/') {
			++i;
		}

		ret_val = g_strdup (work_path + (i - 1));
	} else {
		ret_val = uri_get_path_from_relative (path);
	}
	
	g_free (work_path);

	return ret_val;
}

static gchar *
uri_get_file_uri_from_ghelp_uri (const gchar *path)
{
	char *tmp_path = uri_get_path_from_ghelp_uri (path);
	char *result = g_strconcat ("file://", tmp_path, NULL);
	
	g_free (tmp_path);
	return result;
}
